from __future__ import annotations

import gzip
from contextlib import suppress
from itertools import pairwise
from typing import TYPE_CHECKING

from pytest import mark, param, raises

from utilities.atomicwrites import (
    _MoveDirectoryExistsError,
    _MoveFileExistsError,
    _MoveSourceNotFoundError,
    _WriterDirectoryExistsError,
    _WriterFileExistsError,
    _WriterTemporaryPathEmptyError,
    move,
    move_many,
    writer,
)

if TYPE_CHECKING:
    from pathlib import Path


class TestMove:
    @mark.parametrize("overwrite", [param(True), param(False)])
    def test_file_destination_does_not_exist(
        self, *, tmp_path: Path, overwrite: bool
    ) -> None:
        source = tmp_path.joinpath("source")
        _ = source.write_text("text")
        destination = tmp_path.joinpath("destination")
        move(source, destination, overwrite=overwrite)
        assert destination.is_file()
        assert destination.read_text() == "text"

    def test_file_destination_file_exists(self, *, tmp_path: Path) -> None:
        source = tmp_path.joinpath("source")
        _ = source.write_text("source")
        destination = tmp_path.joinpath("destination")
        _ = destination.write_text("destination")
        move(source, destination, overwrite=True)
        assert destination.is_file()
        assert destination.read_text() == "source"

    def test_file_destination_directory_exists(self, *, tmp_path: Path) -> None:
        source = tmp_path.joinpath("source")
        _ = source.write_text("source")
        destination = tmp_path.joinpath("destination")
        destination.mkdir()
        move(source, destination, overwrite=True)
        assert destination.is_file()
        assert destination.read_text() == "source"

    @mark.parametrize("overwrite", [param(True), param(False)])
    def test_directory_destination_does_not_exist(
        self, *, tmp_path: Path, overwrite: bool
    ) -> None:
        source = tmp_path.joinpath("source")
        source.mkdir()
        source.joinpath("file").touch()
        destination = tmp_path.joinpath("destination")
        move(source, destination, overwrite=overwrite)
        assert destination.is_dir()
        assert len(list(destination.iterdir())) == 1

    def test_directory_destination_file_exists(self, *, tmp_path: Path) -> None:
        source = tmp_path.joinpath("source")
        source.mkdir()
        source.joinpath("file").touch()
        destination = tmp_path.joinpath("destination")
        destination.touch()
        move(source, destination, overwrite=True)
        assert destination.is_dir()
        assert len(list(destination.iterdir())) == 1

    def test_directory_destination_directory_exists(self, *, tmp_path: Path) -> None:
        source = tmp_path.joinpath("source")
        source.mkdir()
        source.joinpath("file").touch()
        destination = tmp_path.joinpath("destination")
        destination.mkdir()
        for i in range(2):
            destination.joinpath(f"file{i}").touch()
        move(source, destination, overwrite=True)
        assert destination.is_dir()
        assert len(list(destination.iterdir())) == 1

    @mark.parametrize("overwrite", [param(True), param(False)])
    def test_error_source_not_found(self, *, tmp_path: Path, overwrite: bool) -> None:
        with raises(_MoveSourceNotFoundError, match="Source '.*' does not exist"):
            move(
                tmp_path.joinpath("source"),
                tmp_path.joinpath("destination"),
                overwrite=overwrite,
            )

    def test_error_file_exists(self, *, tmp_path: Path) -> None:
        source = tmp_path.joinpath("source")
        source.touch()
        destination = tmp_path.joinpath("destination")
        destination.touch()
        with raises(
            _MoveFileExistsError,
            match="Cannot move file '.*' as destination '.*' already exists",
        ):
            move(source, destination)

    def test_error_directory_exists(self, *, tmp_path: Path) -> None:
        source = tmp_path.joinpath("source")
        source.mkdir()
        destination = tmp_path.joinpath("destination")
        destination.touch()
        with raises(
            _MoveDirectoryExistsError,
            match="Cannot move directory '.*' as destination '.*' already exists",
        ):
            move(source, destination)


class TestMoveMany:
    def test_many(self, *, tmp_path: Path) -> None:
        n = 5
        files = [tmp_path.joinpath(f"file{i}") for i in range(n + 2)]
        for i, file in enumerate(files[:-1]):
            _ = file.write_text(str(i))
        move_many(*pairwise(files), overwrite=True)
        for i, file in enumerate(files[1:], start=1):
            assert file.read_text() == str(i - 1)


class TestWriter:
    def test_main(self, *, tmp_path: Path) -> None:
        path = tmp_path.joinpath("file.txt")
        with writer(path) as temp:
            _ = temp.write_text("contents")
        assert path.is_file()
        assert path.read_text() == "contents"

    def test_gzip(self, *, tmp_path: Path) -> None:
        path = tmp_path.joinpath("file.txt")
        with writer(path, compress=True) as temp:
            _ = temp.write_bytes(b"contents")
        assert path.is_file()
        with gzip.open(path) as gz:
            assert gz.read() == b"contents"

    def test_error_temporary_path_empty(self, *, tmp_path: Path) -> None:
        with (
            raises(
                _WriterTemporaryPathEmptyError, match="Temporary path '.*' is empty"
            ),
            writer(tmp_path),
        ):
            pass

    def test_error_file_exists(self, *, tmp_path: Path) -> None:
        path = tmp_path.joinpath("file.txt")
        path.touch()
        with (
            raises(
                _WriterFileExistsError,
                match="Cannot write to '.*' as file already exists",
            ),
            writer(path) as temp,
        ):
            _ = temp.write_text("new contents")

    def test_error_directory_exists(self, *, tmp_path: Path) -> None:
        path = tmp_path.joinpath("dir")
        path.mkdir()
        with (
            raises(
                _WriterDirectoryExistsError,
                match="Cannot write to '.*' as directory already exists",
            ),
            writer(path) as temp,
        ):
            temp.mkdir()

    @mark.parametrize(
        ("error", "expected"),
        [param(KeyboardInterrupt, False), param(ValueError, True)],
    )
    def test_error_during_write(
        self, *, tmp_path: Path, error: type[Exception], expected: bool
    ) -> None:
        path = tmp_path.joinpath("file.txt")

        def raise_error() -> None:
            raise error

        with writer(path) as temp, suppress(Exception):
            _ = temp.write_text("contents")
            raise_error()
        is_non_empty = len(list(tmp_path.iterdir())) >= 1
        assert is_non_empty is expected



from __future__ import annotations

import re
from contextlib import contextmanager
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator
    from types import TracebackType


class NoOpContextManager:
    """Context-manager for no-op."""

    def __enter__(self) -> None:
        return None

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        _ = (exc_type, exc_val, traceback)
        return False


##


_SUPER_OBJECT_HAS_NO_ATTRIBUTE = re.compile(r"'super' object has no attribute '\w+'")


@contextmanager
def suppress_super_object_attribute_error() -> Iterator[None]:
    """Suppress the super() attribute error, for mix-ins."""
    try:
        yield
    except AttributeError as error:
        if not _SUPER_OBJECT_HAS_NO_ATTRIBUTE.search(error.args[0]):
            raise


__all__ = ["NoOpContextManager", "suppress_super_object_attribute_error"]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from platform import system
from re import sub
from typing import TYPE_CHECKING, Literal, assert_never, override

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

System = Literal["windows", "mac", "linux"]


def get_system() -> System:
    """Get the system/OS name."""
    sys = system()
    if sys == "Windows":  # skipif-not-windows
        return "windows"
    if sys == "Darwin":  # skipif-not-macos
        return "mac"
    if sys == "Linux":  # skipif-not-linux
        return "linux"
    raise GetSystemError(sys=sys)  # pragma: no cover


@dataclass(kw_only=True, slots=True)
class GetSystemError(Exception):
    sys: str

    @override
    def __str__(self) -> str:
        return (  # pragma: no cover
            f"System must be one of Windows, Darwin, Linux; got {self.sys!r} instead"
        )


SYSTEM = get_system()
IS_WINDOWS = SYSTEM == "windows"
IS_MAC = SYSTEM == "mac"
IS_LINUX = SYSTEM == "linux"
IS_NOT_WINDOWS = not IS_WINDOWS
IS_NOT_MAC = not IS_MAC
IS_NOT_LINUX = not IS_LINUX


##


def get_max_pid() -> int | None:
    """Get the maximum process ID."""
    match SYSTEM:
        case "windows":  # pragma: no cover
            return None
        case "mac":  # skipif-not-macos
            return 99999
        case "linux":  # skipif-not-linux
            path = Path("/proc/sys/kernel/pid_max")
            try:
                return int(path.read_text())
            except FileNotFoundError:  # pragma: no cover
                return None
        case _ as never:
            assert_never(never)


MAX_PID = get_max_pid()


##


def get_strftime(text: str, /) -> str:
    """Get a platform-specific format string."""
    match SYSTEM:
        case "windows":  # skipif-not-windows
            return text
        case "mac":  # skipif-not-macos
            return text
        case "linux":  # skipif-not-linux
            return sub("%Y", "%4Y", text)
        case _ as never:
            assert_never(never)


##


def maybe_yield_lower_case(text: Iterable[str], /) -> Iterator[str]:
    """Yield lower-cased text if the platform is case-insentive."""
    match SYSTEM:
        case "windows":  # skipif-not-windows
            yield from (t.lower() for t in text)
        case "mac":  # skipif-not-macos
            yield from (t.lower() for t in text)
        case "linux":  # skipif-not-linux
            yield from text
        case _ as never:
            assert_never(never)


__all__ = [
    "IS_LINUX",
    "IS_MAC",
    "IS_NOT_LINUX",
    "IS_NOT_MAC",
    "IS_NOT_WINDOWS",
    "IS_WINDOWS",
    "MAX_PID",
    "SYSTEM",
    "GetSystemError",
    "System",
    "get_max_pid",
    "get_strftime",
    "get_system",
    "maybe_yield_lower_case",
]



from __future__ import annotations

from asyncio import Event, Queue, run
from collections import deque
from collections.abc import ItemsView, KeysView, ValuesView
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from itertools import chain
from re import search
from typing import TYPE_CHECKING, ClassVar, Self

from hypothesis import assume, given
from hypothesis.strategies import (
    DataObject,
    booleans,
    data,
    dictionaries,
    integers,
    lists,
    none,
    permutations,
    sampled_from,
)
from pytest import RaisesGroup, raises

from utilities.asyncio import (
    AsyncDict,
    EnhancedQueue,
    EnhancedTaskGroup,
    UniquePriorityQueue,
    UniqueQueue,
    get_event,
    get_items,
    get_items_nowait,
    put_items,
    put_items_nowait,
    sleep_max,
    sleep_rounded,
    sleep_td,
    sleep_until,
    stream_command,
    timeout_td,
)
from utilities.dataclasses import replace_non_sentinel
from utilities.hypothesis import pairs, sentinels, text_ascii
from utilities.iterables import unique_everseen
from utilities.pytest import skipif_windows
from utilities.sentinel import Sentinel, sentinel
from utilities.timer import Timer
from utilities.whenever import MILLISECOND, SECOND, get_now

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from whenever import TimeDelta

    from utilities.types import MaybeCallableEvent

async_dicts = dictionaries(text_ascii(), integers()).map(AsyncDict)


class TestAsyncDict:
    @given(dict_=async_dicts)
    async def test_aenter(self, *, dict_: AsyncDict[str, int]) -> None:
        async with dict_:
            ...

    @given(dict_=async_dicts)
    async def test_clear(self, *, dict_: AsyncDict[str, int]) -> None:
        await dict_.clear()
        assert len(dict_) == 0

    @given(dict_=async_dicts, key=text_ascii())
    def test_contains(self, *, dict_: AsyncDict[str, int], key: str) -> None:
        assert isinstance(key in dict_, bool)

    @given(dict_=async_dicts)
    def test_copy(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(dict_.copy(), AsyncDict)

    @given(dict_=async_dicts, key=text_ascii())
    async def test_del(self, *, dict_: AsyncDict[str, int], key: str) -> None:
        if key in dict_:
            await dict_.del_(key)
        else:
            with raises(KeyError):
                await dict_.del_(key)

    @given(dict_=async_dicts)
    def test_empty(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(dict_.empty, bool)

    @given(dicts=pairs(async_dicts))
    def test_eq(
        self, *, dicts: tuple[AsyncDict[str, int], AsyncDict[str, int]]
    ) -> None:
        first, second = dicts
        assert isinstance(first == second, bool)

    @given(keys=lists(text_ascii()))
    def test_fromkeys(self, *, keys: list[str]) -> None:
        dict_ = AsyncDict.fromkeys(keys)
        assert isinstance(dict_, AsyncDict)

    @given(dict_=async_dicts, key=text_ascii())
    def test_get(self, *, dict_: AsyncDict[str, int], key: str) -> None:
        if key in dict_:
            assert isinstance(dict_.get(key), int)
        else:
            assert dict_.get(key) is None

    @given(dict_=async_dicts, key=text_ascii())
    def test_get_default(self, *, dict_: AsyncDict[str, int], key: str) -> None:
        value = dict_.get(key, None)
        assert isinstance(value, int) or (value is None)

    @given(dict_=async_dicts, key=text_ascii())
    def test_getitem(self, *, dict_: AsyncDict[str, int], key: str) -> None:
        if key in dict_:
            assert isinstance(dict_[key], int)
        else:
            with raises(KeyError):
                _ = dict_[key]

    @given(dict_=async_dicts)
    def test_items(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(dict_.items(), ItemsView)
        for key, value in dict_.items():
            assert isinstance(key, str)
            assert isinstance(value, int)

    @given(dict_=async_dicts)
    def test_iter(self, *, dict_: AsyncDict[str, int]) -> None:
        for key in dict_:
            assert isinstance(key, str)

    @given(dict_=async_dicts)
    def test_keys(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(dict_.keys(), KeysView)
        for key in dict_.keys():  # noqa: SIM118
            assert isinstance(key, str)

    @given(dict_=async_dicts)
    def test_len(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(len(dict_), int)

    @given(dict_=async_dicts, key=text_ascii())
    async def test_pop(self, *, dict_: AsyncDict[str, int], key: str) -> None:
        if key in dict_:
            assert isinstance(await dict_.pop(key), int)
        else:
            with raises(KeyError):
                _ = await dict_.pop(key)

    @given(dict_=async_dicts, key=text_ascii())
    async def test_pop_default(self, *, dict_: AsyncDict[str, int], key: str) -> None:
        value = await dict_.pop(key, None)
        assert isinstance(value, int) or (value is None)

    @given(dict_=async_dicts)
    async def test_popitem(self, *, dict_: AsyncDict[str, int]) -> None:
        if len(dict_) >= 1:
            key, value = await dict_.popitem()
            assert isinstance(key, str)
            assert isinstance(value, int)
        else:
            with raises(KeyError):
                _ = await dict_.popitem()

    @given(dict_=async_dicts)
    def test_repr(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(repr(dict_), str)

    @given(dict_=async_dicts)
    def test_reversed(self, *, dict_: AsyncDict[str, int]) -> None:
        for key in reversed(dict_):
            assert isinstance(key, str)

    @given(dict_=async_dicts, key=text_ascii(), value=integers())
    async def test_set(
        self, *, dict_: AsyncDict[str, int], key: str, value: int
    ) -> None:
        await dict_.set(key, value)

    @given(dict_=async_dicts, key=text_ascii(), value=integers())
    async def test_setdefault(
        self, *, dict_: AsyncDict[str, int], key: str, value: int
    ) -> None:
        assert isinstance(await dict_.setdefault(key, value), int)

    @given(dict_=async_dicts)
    def test_str(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(str(dict_), str)

    @given(dicts=pairs(async_dicts))
    async def test_update(
        self, *, dicts: tuple[AsyncDict[str, int], AsyncDict[str, int]]
    ) -> None:
        first, second = dicts
        await first.update(second)

    @given(dict_=async_dicts)
    def test_values(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(dict_.values(), ValuesView)
        for value in dict_.values():
            assert isinstance(value, int)


class TestEnhancedQueue:
    @given(
        xs=lists(integers()),
        wait=booleans(),
        put_all=booleans(),
        get_reverse=booleans(),
    )
    async def test_left(
        self, *, xs: list[int], wait: int, put_all: bool, get_reverse: bool
    ) -> None:
        _ = assume(not ((len(xs) == 0) and wait))
        deq: deque[int] = deque()
        for x in xs:
            deq.appendleft(x)
        queue: EnhancedQueue[int] = EnhancedQueue()
        if put_all:
            if wait:
                await queue.put_left(*xs)
            else:
                queue.put_left_nowait(*xs)
        else:
            for i, x in enumerate(xs, start=1):
                if wait:
                    await queue.put_left(x)
                else:
                    queue.put_left_nowait(x)
                assert queue.qsize() == i
        assert list(deq) == xs[::-1]
        if wait:
            res = await queue.get_all(reverse=get_reverse)
        else:
            res = queue.get_all_nowait(reverse=get_reverse)
        expected = xs if get_reverse else xs[::-1]
        assert res == expected

    @given(
        xs=lists(integers()),
        wait=booleans(),
        put_all=booleans(),
        get_reverse=booleans(),
    )
    async def test_right(
        self, *, xs: list[int], wait: int, put_all: bool, get_reverse: bool
    ) -> None:
        _ = assume(not ((len(xs) == 0) and wait))
        deq: deque[int] = deque()
        for x in xs:
            deq.append(x)
        queue: EnhancedQueue[int] = EnhancedQueue()
        if put_all:
            if wait:
                await queue.put_right(*xs)
            else:
                queue.put_right_nowait(*xs)
            assert queue.qsize() == len(xs)
        else:
            for i, x in enumerate(xs, start=1):
                if wait:
                    await queue.put_right(x)
                else:
                    queue.put_right_nowait(x)
                assert queue.qsize() == i
        assert list(deq) == xs
        if wait:
            res = await queue.get_all(reverse=get_reverse)
        else:
            res = queue.get_all_nowait(reverse=get_reverse)
        expected = xs[::-1] if get_reverse else xs
        assert res == expected


class TestEnhancedTaskGroup:
    delta: ClassVar[TimeDelta] = 0.05 * SECOND

    async def test_create_task_context_coroutine(self) -> None:
        flag: bool = False

        @asynccontextmanager
        async def yield_true() -> AsyncIterator[None]:
            nonlocal flag
            try:
                flag = True
                yield
            finally:
                flag = False

        assert not flag
        async with EnhancedTaskGroup(timeout=2 * self.delta) as tg:
            _ = tg.create_task_context(yield_true())
            await sleep_td(self.delta)
            assert flag
        assert not flag

    async def test_max_tasks_disabled(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup() as tg:
                for _ in range(10):
                    _ = tg.create_task(sleep_td(self.delta))
        assert timer <= 2 * self.delta

    async def test_max_tasks_enabled(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup(max_tasks=2) as tg:
                for _ in range(10):
                    _ = tg.create_task(sleep_td(self.delta))
        assert timer >= 5 * self.delta

    async def test_run_or_create_many_tasks_parallel_with_max_tasks_two(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup(max_tasks=2) as tg:
                assert not tg._is_debug()
                for _ in range(10):
                    _ = await tg.run_or_create_many_tasks(sleep_td, self.delta)
        assert timer >= 5 * self.delta

    async def test_run_or_create_many_tasks_serial_with_debug(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup(debug=True) as tg:
                assert tg._is_debug()
                for _ in range(10):
                    _ = await tg.run_or_create_many_tasks(sleep_td, self.delta)
        assert timer >= 10 * self.delta

    async def test_run_or_create_task_parallel_with_max_tasks_none(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup() as tg:
                assert not tg._is_debug()
                for _ in range(10):
                    _ = await tg.run_or_create_task(sleep_td(self.delta))
        assert timer <= 2 * self.delta

    async def test_run_or_create_task_parallel_with_max_tasks_two(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup(max_tasks=2) as tg:
                assert not tg._is_debug()
                for _ in range(10):
                    _ = await tg.run_or_create_task(sleep_td(self.delta))
        assert timer >= 5 * self.delta

    async def test_run_or_create_task_serial_with_max_tasks_negative(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup(max_tasks=-1) as tg:
                assert tg._is_debug()
                for _ in range(10):
                    _ = await tg.run_or_create_task(sleep_td(self.delta))
        assert timer >= 10 * self.delta

    async def test_run_or_create_task_serial_with_debug(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup(debug=True) as tg:
                assert tg._is_debug()
                for _ in range(10):
                    _ = await tg.run_or_create_task(sleep_td(self.delta))
        assert timer >= 10 * self.delta

    async def test_timeout_pass(self) -> None:
        async with EnhancedTaskGroup(timeout=2 * self.delta) as tg:
            _ = tg.create_task(sleep_td(self.delta))

    async def test_timeout_fail(self) -> None:
        with RaisesGroup(TimeoutError):
            async with EnhancedTaskGroup(timeout=self.delta) as tg:
                _ = tg.create_task(sleep_td(2 * self.delta))

    async def test_custom_error(self) -> None:
        class CustomError(Exception): ...

        with RaisesGroup(CustomError):
            async with EnhancedTaskGroup(timeout=self.delta, error=CustomError) as tg:
                _ = tg.create_task(sleep_td(2 * self.delta))


class TestGetEvent:
    def test_event(self) -> None:
        event = Event()
        assert get_event(event=event) is event

    @given(event=none() | sentinels())
    def test_none_or_sentinel(self, *, event: None | Sentinel) -> None:
        assert get_event(event=event) is event

    def test_replace_non_sentinel(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            event: Event = field(default_factory=Event)

            def replace(
                self, *, event: MaybeCallableEvent | Sentinel = sentinel
            ) -> Self:
                return replace_non_sentinel(self, event=get_event(event=event))

        event1, event2, event3 = Event(), Event(), Event()
        obj = Example(event=event1)
        assert obj.event is event1
        assert obj.replace().event is event1
        assert obj.replace(event=event2).event is event2
        assert obj.replace(event=lambda: event3).event is event3

    def test_callable(self) -> None:
        event = Event()
        assert get_event(event=lambda: event) is event


class TestGetItems:
    @given(
        xs=lists(integers(), min_size=1),
        max_size=integers(1, 10) | none(),
        wait=booleans(),
    )
    async def test_main(
        self, *, xs: list[int], max_size: int | None, wait: bool
    ) -> None:
        queue: Queue[int] = Queue()
        put_items_nowait(xs, queue)
        if wait:
            result = await get_items(queue, max_size=max_size)
        else:
            result = get_items_nowait(queue, max_size=max_size)
        assert result == xs[:max_size]


class TestPutItems:
    @given(xs=lists(integers(), min_size=1), wait=booleans())
    async def test_main(self, *, xs: list[int], wait: bool) -> None:
        queue: Queue[int] = Queue()
        if wait:
            put_items_nowait(xs, queue)
        else:
            await put_items(xs, queue)
        result: list[int] = []
        while not queue.empty():
            result.append(await queue.get())
        assert result == xs


class TestUniquePriorityQueue:
    @given(data=data(), texts=lists(text_ascii(min_size=1), min_size=1, unique=True))
    async def test_main(self, *, data: DataObject, texts: list[str]) -> None:
        items = list(enumerate(texts))
        extra = data.draw(lists(sampled_from(items)))
        items_use = data.draw(permutations(list(chain(items, extra))))
        queue: UniquePriorityQueue[int, str] = UniquePriorityQueue()
        assert queue._set == set()
        for item in items_use:
            await queue.put(item)
        assert queue._set == set(texts)
        result = await get_items(queue)
        assert result == items
        assert queue._set == set()


class TestUniqueQueue:
    @given(x=lists(integers(), min_size=1))
    async def test_main(self, *, x: list[int]) -> None:
        queue: UniqueQueue[int] = UniqueQueue()
        assert queue._set == set()
        for x_i in x:
            await queue.put(x_i)
        assert queue._set == set(x)
        result = await get_items(queue)
        expected = list(unique_everseen(x))
        assert result == expected
        assert queue._set == set()


class TestSleepMaxDur:
    delta: ClassVar[TimeDelta] = 0.05 * SECOND

    async def test_main(self) -> None:
        with Timer() as timer:
            await sleep_max(self.delta)
        assert timer <= 2 * self.delta

    async def test_none(self) -> None:
        with Timer() as timer:
            await sleep_max()
        assert timer <= self.delta


class TestSleepTD:
    delta: ClassVar[TimeDelta] = 0.05 * SECOND

    async def test_main(self) -> None:
        with Timer() as timer:
            await sleep_td(self.delta)
        assert timer <= 2 * self.delta

    async def test_none(self) -> None:
        with Timer() as timer:
            await sleep_td()
        assert timer <= self.delta


class TestSleepUntil:
    async def test_main(self) -> None:
        await sleep_until(get_now() + 0.05 * SECOND)


class TestSleepUntilRounded:
    async def test_main(self) -> None:
        await sleep_rounded(10 * MILLISECOND)


class TestStreamCommand:
    delta: ClassVar[TimeDelta] = 0.05 * SECOND

    @skipif_windows
    async def test_main(self) -> None:
        output = await stream_command(
            'echo "stdout message" && sleep 0.1 && echo "stderr message" >&2'
        )
        await sleep_td(self.delta)
        assert output.return_code == 0
        assert output.stdout == "stdout message\n"
        assert output.stderr == "stderr message\n"

    @skipif_windows
    async def test_error(self) -> None:
        output = await stream_command("this-is-an-error")
        await sleep_td(self.delta)
        assert output.return_code == 127
        assert output.stdout == ""
        assert search(
            r"^/bin/sh: (1: )?this-is-an-error: (command )?not found$", output.stderr
        )


class TestTimeoutTD:
    delta: ClassVar[TimeDelta] = 0.05 * SECOND

    async def test_pass(self) -> None:
        async with timeout_td(2 * self.delta):
            await sleep_td(self.delta)

    async def test_fail(self) -> None:
        with raises(TimeoutError):
            async with timeout_td(self.delta):
                await sleep_td(2 * self.delta)

    async def test_custom_error(self) -> None:
        class CustomError(Exception): ...

        with raises(CustomError):
            async with timeout_td(self.delta, error=CustomError):
                await sleep_td(2 * self.delta)


if __name__ == "__main__":
    _ = run(
        stream_command('echo "stdout message" && sleep 2 && echo "stderr message" >&2')
    )



from __future__ import annotations

from collections.abc import Sized
from typing import assert_never, cast

from inflect import Word, engine

_ENGINE = engine()


def counted_noun(obj: int | Sized, noun: str, /) -> str:
    """Construct a counted noun."""
    match obj:
        case int() as count:
            ...
        case Sized() as sized:
            count = len(sized)
        case _ as never:
            assert_never(never)
    word = cast("Word", noun)
    sin_or_plu = _ENGINE.plural_noun(word, count=count)
    return f"{count} {sin_or_plu}"


__all__ = ["counted_noun"]



from __future__ import annotations

import datetime as dt
from dataclasses import dataclass, field
from logging import DEBUG
from typing import TYPE_CHECKING, ClassVar, Self
from zoneinfo import ZoneInfo

from hypothesis import given
from hypothesis.strategies import integers, none, sampled_from, timezones
from pytest import mark, param, raises
from whenever import (
    Date,
    DateDelta,
    DateTimeDelta,
    PlainDateTime,
    Time,
    TimeDelta,
    TimeZoneNotFoundError,
    Weekday,
    YearMonth,
    ZonedDateTime,
)

from utilities.dataclasses import replace_non_sentinel
from utilities.hypothesis import (
    assume_does_not_raise,
    date_deltas,
    dates,
    pairs,
    plain_datetimes,
    sentinels,
    times,
    zoned_datetimes,
)
from utilities.sentinel import Sentinel, sentinel
from utilities.tzdata import HongKong, Tokyo
from utilities.tzlocal import LOCAL_TIME_ZONE_NAME
from utilities.whenever import (
    DATE_DELTA_MAX,
    DATE_DELTA_MIN,
    DATE_DELTA_PARSABLE_MAX,
    DATE_DELTA_PARSABLE_MIN,
    DATE_TIME_DELTA_MAX,
    DATE_TIME_DELTA_MIN,
    DATE_TIME_DELTA_PARSABLE_MAX,
    DATE_TIME_DELTA_PARSABLE_MIN,
    DAY,
    MICROSECOND,
    MINUTE,
    MONTH,
    NOW_LOCAL,
    NOW_UTC,
    SECOND,
    TIME_DELTA_MAX,
    TIME_DELTA_MIN,
    TODAY_LOCAL,
    TODAY_UTC,
    ZERO_DAYS,
    ZONED_DATE_TIME_MAX,
    ZONED_DATE_TIME_MIN,
    MeanDateTimeError,
    MinMaxDateError,
    ToMonthsAndDaysError,
    ToNanosecondsError,
    ToPyTimeDeltaError,
    WheneverLogRecord,
    _MinMaxDateMaxDateError,
    _MinMaxDateMinDateError,
    _MinMaxDatePeriodError,
    _RoundDateOrDateTimeDateTimeIntraDayWithWeekdayError,
    _RoundDateOrDateTimeDateWithIntradayDeltaError,
    _RoundDateOrDateTimeDateWithWeekdayError,
    _RoundDateOrDateTimeIncrementError,
    _RoundDateOrDateTimeInvalidDurationError,
    _ToDaysMonthsError,
    _ToDaysNanosecondsError,
    _ToHoursMonthsError,
    _ToHoursNanosecondsError,
    _ToMicrosecondsMonthsError,
    _ToMicrosecondsNanosecondsError,
    _ToMillisecondsMonthsError,
    _ToMillisecondsNanosecondsError,
    _ToMinutesMonthsError,
    _ToMinutesNanosecondsError,
    _ToMonthsDaysError,
    _ToMonthsTimeError,
    _ToSecondsMonthsError,
    _ToSecondsNanosecondsError,
    _ToWeeksDaysError,
    _ToWeeksMonthsError,
    _ToWeeksNanosecondsError,
    _ToYearsDaysError,
    _ToYearsMonthsError,
    _ToYearsTimeError,
    add_year_month,
    datetime_utc,
    diff_year_month,
    format_compact,
    from_timestamp,
    from_timestamp_millis,
    from_timestamp_nanos,
    get_now,
    get_now_local,
    get_today,
    get_today_local,
    mean_datetime,
    min_max_date,
    round_date_or_date_time,
    sub_year_month,
    to_date,
    to_date_time_delta,
    to_days,
    to_hours,
    to_local_plain,
    to_microseconds,
    to_milliseconds,
    to_minutes,
    to_months,
    to_months_and_days,
    to_nanoseconds,
    to_py_date_or_date_time,
    to_py_time_delta,
    to_seconds,
    to_time_delta,
    to_weeks,
    to_years,
    to_zoned_date_time,
    two_digit_year_month,
)
from utilities.zoneinfo import UTC

if TYPE_CHECKING:
    from collections.abc import Callable

    from _pytest.mark import ParameterSet

    from utilities.sentinel import Sentinel
    from utilities.types import (
        DateOrDateTimeDelta,
        DateTimeRoundMode,
        Delta,
        MaybeCallableDate,
        MaybeCallableZonedDateTime,
        TimeOrDateTimeDelta,
    )


class TestAddAndSubYearMonth:
    x: ClassVar[YearMonth] = YearMonth(2005, 7)
    cases: ClassVar[list[tuple[int, int, YearMonth, YearMonth]]] = [
        (1, 0, YearMonth(2006, 7), YearMonth(2004, 7)),
        (0, 11, YearMonth(2006, 6), YearMonth(2004, 8)),
        (0, 6, YearMonth(2006, 1), YearMonth(2005, 1)),
        (0, 2, YearMonth(2005, 9), YearMonth(2005, 5)),
        (0, 1, YearMonth(2005, 8), YearMonth(2005, 6)),
        (0, 0, YearMonth(2005, 7), YearMonth(2005, 7)),
        (0, -1, YearMonth(2005, 6), YearMonth(2005, 8)),
        (0, -2, YearMonth(2005, 5), YearMonth(2005, 9)),
        (0, -6, YearMonth(2005, 1), YearMonth(2006, 1)),
        (0, -11, YearMonth(2004, 8), YearMonth(2006, 6)),
        (-1, 0, YearMonth(2004, 7), YearMonth(2006, 7)),
    ]

    @mark.parametrize(
        ("years", "months", "expected"), [param(y, m, e) for y, m, e, _ in cases]
    )
    def test_add(self, *, years: int, months: int, expected: YearMonth) -> None:
        result = add_year_month(self.x, years=years, months=months)
        assert result == expected

    @mark.parametrize(
        ("years", "months", "expected"), [param(y, m, e) for y, m, _, e in cases]
    )
    def test_sub(self, *, years: int, months: int, expected: YearMonth) -> None:
        result = sub_year_month(self.x, years=years, months=months)
        assert result == expected


class TestDatetimeUTC:
    @given(datetime=zoned_datetimes())
    def test_main(self, *, datetime: ZonedDateTime) -> None:
        result = datetime_utc(
            datetime.year,
            datetime.month,
            datetime.day,
            hour=datetime.hour,
            minute=datetime.minute,
            second=datetime.second,
            nanosecond=datetime.nanosecond,
        )
        assert result == datetime


class TestDiffYearMonth:
    x: ClassVar[YearMonth] = YearMonth(2005, 7)
    cases: ClassVar[list[ParameterSet]] = [
        param(YearMonth(2004, 7), 1, 0),
        param(YearMonth(2004, 8), 0, 11),
        param(YearMonth(2005, 1), 0, 6),
        param(YearMonth(2005, 5), 0, 2),
        param(YearMonth(2005, 6), 0, 1),
        param(YearMonth(2005, 7), 0, 0),
        param(YearMonth(2005, 8), 0, -1),
        param(YearMonth(2005, 9), 0, -2),
        param(YearMonth(2006, 1), 0, -6),
        param(YearMonth(2006, 6), 0, -11),
        param(YearMonth(2006, 7), -1, 0),
    ]

    @mark.parametrize(("y", "year", "month"), cases)
    def test_main(self, *, y: YearMonth, year: int, month: int) -> None:
        result = diff_year_month(self.x, y)
        expected = 12 * year + month
        assert result == expected

    @mark.parametrize(("y", "year", "month"), cases)
    def test_year_and_month(self, *, y: YearMonth, year: int, month: int) -> None:
        result = diff_year_month(self.x, y, years=True)
        expected = (year, month)
        assert result == expected


class TestFormatCompact:
    @given(date=dates())
    def test_date(self, *, date: Date) -> None:
        result = format_compact(date)
        assert isinstance(result, str)
        parsed = Date.parse_common_iso(result)
        assert parsed == date

    @given(time=times())
    def test_time(self, *, time: Time) -> None:
        result = format_compact(time)
        assert isinstance(result, str)
        parsed = Time.parse_common_iso(result)
        assert parsed.nanosecond == 0
        expected = time.round()
        assert parsed == expected

    @given(datetime=plain_datetimes())
    def test_plain_datetime(self, *, datetime: PlainDateTime) -> None:
        result = format_compact(datetime)
        assert isinstance(result, str)
        parsed = PlainDateTime.parse_common_iso(result)
        assert parsed.nanosecond == 0
        expected = datetime.round()
        assert parsed == expected

    @given(datetime=zoned_datetimes())
    def test_zoned_datetime(self, *, datetime: ZonedDateTime) -> None:
        result = format_compact(datetime)
        assert isinstance(result, str)
        parsed = ZonedDateTime.parse_common_iso(result)
        assert parsed.nanosecond == 0
        expected = datetime.round()
        assert parsed == expected


class TestFromTimeStamp:
    @given(
        datetime=zoned_datetimes(time_zone=timezones()).map(lambda d: d.round("second"))
    )
    def test_main(self, *, datetime: ZonedDateTime) -> None:
        timestamp = datetime.timestamp()
        result = from_timestamp(timestamp, time_zone=ZoneInfo(datetime.tz))
        assert result == datetime

    @given(
        datetime=zoned_datetimes(time_zone=timezones()).map(
            lambda d: d.round("millisecond")
        )
    )
    def test_millis(self, *, datetime: ZonedDateTime) -> None:
        timestamp = datetime.timestamp_millis()
        result = from_timestamp_millis(timestamp, time_zone=ZoneInfo(datetime.tz))
        assert result == datetime

    @given(datetime=zoned_datetimes(time_zone=timezones()))
    def test_nanos(self, *, datetime: ZonedDateTime) -> None:
        timestamp = datetime.timestamp_nanos()
        result = from_timestamp_nanos(timestamp, time_zone=ZoneInfo(datetime.tz))
        assert result == datetime


class TestGetNow:
    @given(time_zone=timezones())
    def test_function(self, *, time_zone: ZoneInfo) -> None:
        with assume_does_not_raise(TimeZoneNotFoundError):
            now = get_now(time_zone=time_zone)
        assert isinstance(now, ZonedDateTime)
        assert now.tz == time_zone.key

    def test_constant(self) -> None:
        assert isinstance(NOW_UTC, ZonedDateTime)
        assert NOW_UTC.tz == "UTC"


class TestGetNowLocal:
    def test_function(self) -> None:
        now = get_now_local()
        assert isinstance(now, ZonedDateTime)
        ETC = ZoneInfo("Etc/UTC")  # noqa: N806
        time_zones = {ETC, HongKong, Tokyo, UTC}
        assert any(now.tz == time_zone.key for time_zone in time_zones)

    def test_constant(self) -> None:
        assert isinstance(NOW_LOCAL, ZonedDateTime)
        assert NOW_LOCAL.tz == LOCAL_TIME_ZONE_NAME


class TestGetToday:
    def test_function(self) -> None:
        today = get_today()
        assert isinstance(today, Date)

    def test_constant(self) -> None:
        assert isinstance(TODAY_UTC, Date)


class TestGetTodayLocal:
    def test_function(self) -> None:
        today = get_today_local()
        assert isinstance(today, Date)

    def test_constant(self) -> None:
        assert isinstance(TODAY_LOCAL, Date)


class TestMeanDateTime:
    threshold: ClassVar[TimeDelta] = 100 * MICROSECOND

    @given(datetime=zoned_datetimes())
    def test_one(self, *, datetime: ZonedDateTime) -> None:
        result = mean_datetime([datetime])
        assert result == datetime

    @given(datetime=zoned_datetimes())
    def test_many(self, *, datetime: ZonedDateTime) -> None:
        result = mean_datetime([datetime, datetime + MINUTE])
        expected = datetime + 30 * SECOND
        assert abs(result - expected) <= self.threshold

    @given(datetime=zoned_datetimes())
    def test_weights(self, *, datetime: ZonedDateTime) -> None:
        result = mean_datetime([datetime, datetime + MINUTE], weights=[1, 3])
        expected = datetime + 45 * SECOND
        assert abs(result - expected) <= self.threshold

    def test_error(self) -> None:
        with raises(MeanDateTimeError, match="Mean requires at least 1 datetime"):
            _ = mean_datetime([])


class TestMinMax:
    def test_date_delta_min(self) -> None:
        with raises(ValueError, match="Addition result out of bounds"):
            _ = DATE_DELTA_MIN - DateDelta(days=1)

    def test_date_delta_max(self) -> None:
        with raises(ValueError, match="Addition result out of bounds"):
            _ = DATE_DELTA_MAX + DateDelta(days=1)

    def test_date_delta_parsable_min(self) -> None:
        self._format_parse_date_delta(DATE_DELTA_PARSABLE_MIN)
        with raises(ValueError, match="Invalid format: '.*'"):
            self._format_parse_date_delta(DATE_DELTA_PARSABLE_MIN - DateDelta(days=1))

    def test_date_delta_parsable_max(self) -> None:
        self._format_parse_date_delta(DATE_DELTA_PARSABLE_MAX)
        with raises(ValueError, match="Invalid format: '.*'"):
            self._format_parse_date_delta(DATE_DELTA_PARSABLE_MAX + DateDelta(days=1))

    def test_date_time_delta_min(self) -> None:
        nanos = to_nanoseconds(DATE_TIME_DELTA_MIN)
        with raises(ValueError, match="Out of range"):
            _ = to_date_time_delta(nanos - 1)

    def test_date_time_delta_max(self) -> None:
        nanos = to_nanoseconds(DATE_TIME_DELTA_MAX)
        with raises(ValueError, match="Out of range"):
            _ = to_date_time_delta(nanos + 1)

    def test_date_time_delta_parsable_min(self) -> None:
        self._format_parse_date_time_delta(DATE_TIME_DELTA_PARSABLE_MIN)
        nanos = to_nanoseconds(DATE_TIME_DELTA_PARSABLE_MIN)
        with raises(ValueError, match="Invalid format or out of range: '.*'"):
            self._format_parse_date_time_delta(to_date_time_delta(nanos - 1))

    def test_date_time_delta_parsable_max(self) -> None:
        self._format_parse_date_time_delta(DATE_TIME_DELTA_PARSABLE_MAX)
        nanos = to_nanoseconds(DATE_TIME_DELTA_PARSABLE_MAX)
        with raises(ValueError, match="Invalid format or out of range: '.*'"):
            _ = self._format_parse_date_time_delta(to_date_time_delta(nanos + 1))

    def test_plain_date_time_min(self) -> None:
        with raises(ValueError, match=r"Result of subtract\(\) out of range"):
            _ = PlainDateTime.MIN.subtract(nanoseconds=1, ignore_dst=True)

    def test_plain_date_time_max(self) -> None:
        with raises(ValueError, match=r"Result of add\(\) out of range"):
            _ = PlainDateTime.MAX.add(microseconds=1, ignore_dst=True)

    def test_time_delta_min(self) -> None:
        nanos = TIME_DELTA_MIN.in_nanoseconds()
        with raises(ValueError, match="TimeDelta out of range"):
            _ = to_time_delta(nanos - 1)

    def test_time_delta_max(self) -> None:
        nanos = TIME_DELTA_MAX.in_nanoseconds()
        with raises(ValueError, match="TimeDelta out of range"):
            _ = to_time_delta(nanos + 1)

    def test_zoned_date_time_min(self) -> None:
        with raises(ValueError, match="Instant is out of range"):
            _ = ZONED_DATE_TIME_MIN.subtract(nanoseconds=1)

    def test_zoned_date_time_max(self) -> None:
        with raises(ValueError, match="Instant is out of range"):
            _ = ZONED_DATE_TIME_MAX.add(microseconds=1)

    def _format_parse_date_delta(self, delta: DateDelta, /) -> None:
        _ = DateDelta.parse_common_iso(delta.format_common_iso())

    def _format_parse_date_time_delta(self, delta: DateTimeDelta, /) -> None:
        _ = DateTimeDelta.parse_common_iso(delta.format_common_iso())


class TestMinMaxDate:
    @given(
        min_date=dates(max_value=TODAY_LOCAL) | none(),
        max_date=dates(max_value=TODAY_LOCAL) | none(),
        min_age=date_deltas(min_value=ZERO_DAYS) | none(),
        max_age=date_deltas(min_value=ZERO_DAYS) | none(),
    )
    def test_main(
        self,
        *,
        min_date: Date | None,
        max_date: Date | None,
        min_age: DateDelta | None,
        max_age: DateDelta | None,
    ) -> None:
        with (
            assume_does_not_raise(MinMaxDateError),
            assume_does_not_raise(ValueError, match="Resulting date out of range"),
        ):
            min_date_use, max_date_use = min_max_date(
                min_date=min_date, max_date=max_date, min_age=min_age, max_age=max_age
            )
        if (min_date is None) and (max_age is None):
            assert min_date_use is None
        else:
            assert min_date_use is not None
        if (max_date is None) and (min_age is None):
            assert max_date_use is None
        else:
            assert max_date_use is not None
        if min_date_use is not None:
            assert min_date_use <= get_today()
        if max_date_use is not None:
            assert max_date_use <= get_today()
        if (min_date_use is not None) and (max_date_use is not None):
            assert min_date_use <= max_date_use

    @given(date=dates(min_value=TODAY_UTC + DAY))
    def test_error_min_date(self, *, date: Date) -> None:
        with raises(
            _MinMaxDateMinDateError, match="Min date must be at most today; got .* > .*"
        ):
            _ = min_max_date(min_date=date)

    @given(date=dates(min_value=TODAY_UTC + DAY))
    def test_error_max_date(self, *, date: Date) -> None:
        with raises(
            _MinMaxDateMaxDateError, match="Max date must be at most today; got .* > .*"
        ):
            _ = min_max_date(max_date=date)

    @given(dates=pairs(dates(max_value=TODAY_UTC), unique=True, sorted=True))
    def test_error_period(self, *, dates: tuple[Date, Date]) -> None:
        with raises(
            _MinMaxDatePeriodError,
            match="Min date must be at most max date; got .* > .*",
        ):
            _ = min_max_date(min_date=dates[1], max_date=dates[0])


class TestRoundDateOrDateTime:
    @mark.parametrize(
        ("date", "delta", "mode", "expected"),
        [
            param(Date(2000, 1, 1), DateDelta(days=1), "half_even", Date(2000, 1, 1)),
            param(Date(2000, 1, 1), DateDelta(days=2), "half_even", Date(2000, 1, 2)),
            param(Date(2000, 1, 1), DateDelta(days=2), "ceil", Date(2000, 1, 2)),
            param(Date(2000, 1, 1), DateDelta(days=2), "floor", Date(1999, 12, 31)),
            param(Date(2000, 1, 1), DateDelta(days=2), "half_ceil", Date(2000, 1, 2)),
            param(
                Date(2000, 1, 1), DateDelta(days=2), "half_floor", Date(1999, 12, 31)
            ),
            param(Date(2000, 1, 2), DateDelta(days=2), "half_even", Date(2000, 1, 2)),
            param(Date(2000, 1, 2), DateDelta(days=2), "ceil", Date(2000, 1, 2)),
            param(Date(2000, 1, 2), DateDelta(days=2), "floor", Date(2000, 1, 2)),
            param(Date(2000, 1, 2), DateDelta(days=2), "half_ceil", Date(2000, 1, 2)),
            param(Date(2000, 1, 2), DateDelta(days=2), "half_floor", Date(2000, 1, 2)),
            param(Date(2000, 1, 1), DateDelta(days=3), "half_even", Date(2000, 1, 1)),
            param(Date(2000, 1, 1), DateDelta(days=3), "ceil", Date(2000, 1, 1)),
            param(Date(2000, 1, 1), DateDelta(days=3), "floor", Date(2000, 1, 1)),
            param(Date(2000, 1, 1), DateDelta(days=3), "half_ceil", Date(2000, 1, 1)),
            param(Date(2000, 1, 1), DateDelta(days=3), "half_floor", Date(2000, 1, 1)),
            param(Date(2000, 1, 2), DateDelta(days=3), "half_even", Date(2000, 1, 4)),
            param(Date(2000, 1, 2), DateDelta(days=3), "ceil", Date(2000, 1, 4)),
            param(Date(2000, 1, 2), DateDelta(days=3), "floor", Date(2000, 1, 1)),
            param(Date(2000, 1, 2), DateDelta(days=3), "half_ceil", Date(2000, 1, 4)),
            param(Date(2000, 1, 2), DateDelta(days=3), "half_floor", Date(2000, 1, 1)),
            param(Date(2000, 1, 3), DateDelta(days=3), "half_even", Date(2000, 1, 4)),
            param(Date(2000, 1, 3), DateDelta(days=3), "ceil", Date(2000, 1, 4)),
            param(Date(2000, 1, 3), DateDelta(days=3), "floor", Date(2000, 1, 1)),
            param(Date(2000, 1, 3), DateDelta(days=3), "half_ceil", Date(2000, 1, 4)),
            param(Date(2000, 1, 3), DateDelta(days=3), "half_floor", Date(2000, 1, 4)),
        ],
    )
    def test_date_daily(
        self,
        *,
        date: Date,
        delta: Delta,
        mode: DateTimeRoundMode,
        expected: ZonedDateTime,
    ) -> None:
        result = round_date_or_date_time(date, delta, mode=mode)
        assert result == expected

    @mark.parametrize(
        ("date", "weekday", "expected"),
        [
            param(Date(2000, 1, 1), None, Date(1999, 12, 27)),
            param(Date(2000, 1, 2), None, Date(1999, 12, 27)),
            param(Date(2000, 1, 3), None, Date(2000, 1, 3)),
            param(Date(2000, 1, 4), None, Date(2000, 1, 3)),
            param(Date(2000, 1, 5), None, Date(2000, 1, 3)),
            param(Date(2000, 1, 6), None, Date(2000, 1, 3)),
            param(Date(2000, 1, 7), None, Date(2000, 1, 3)),
            param(Date(2000, 1, 8), None, Date(2000, 1, 3)),
            param(Date(2000, 1, 9), None, Date(2000, 1, 3)),
            param(Date(2000, 1, 10), None, Date(2000, 1, 10)),
            param(Date(2000, 1, 11), None, Date(2000, 1, 10)),
            param(Date(2000, 1, 1), Weekday.WEDNESDAY, Date(1999, 12, 29)),
            param(Date(2000, 1, 2), Weekday.WEDNESDAY, Date(1999, 12, 29)),
            param(Date(2000, 1, 3), Weekday.WEDNESDAY, Date(1999, 12, 29)),
            param(Date(2000, 1, 4), Weekday.WEDNESDAY, Date(1999, 12, 29)),
            param(Date(2000, 1, 5), Weekday.WEDNESDAY, Date(2000, 1, 5)),
            param(Date(2000, 1, 6), Weekday.WEDNESDAY, Date(2000, 1, 5)),
            param(Date(2000, 1, 7), Weekday.WEDNESDAY, Date(2000, 1, 5)),
            param(Date(2000, 1, 8), Weekday.WEDNESDAY, Date(2000, 1, 5)),
            param(Date(2000, 1, 9), Weekday.WEDNESDAY, Date(2000, 1, 5)),
            param(Date(2000, 1, 10), Weekday.WEDNESDAY, Date(2000, 1, 5)),
            param(Date(2000, 1, 11), Weekday.WEDNESDAY, Date(2000, 1, 5)),
            param(Date(2000, 1, 12), Weekday.WEDNESDAY, Date(2000, 1, 12)),
            param(Date(2000, 1, 13), Weekday.WEDNESDAY, Date(2000, 1, 12)),
        ],
    )
    def test_date_weekly(
        self, *, date: Date, weekday: Weekday | None, expected: ZonedDateTime
    ) -> None:
        result = round_date_or_date_time(
            date, DateDelta(weeks=1), mode="floor", weekday=weekday
        )
        assert result == expected
        if weekday is not None:
            assert result.day_of_week() is weekday

    @mark.parametrize(
        ("delta", "expected"),
        [
            param(TimeDelta(hours=2), ZonedDateTime(2000, 1, 2, 2, tz=UTC.key)),
            param(TimeDelta(minutes=2), ZonedDateTime(2000, 1, 2, 3, 4, tz=UTC.key)),
            param(TimeDelta(seconds=2), ZonedDateTime(2000, 1, 2, 3, 4, 4, tz=UTC.key)),
            param(
                TimeDelta(milliseconds=2),
                ZonedDateTime(2000, 1, 2, 3, 4, 5, nanosecond=122000000, tz=UTC.key),
            ),
            param(
                TimeDelta(microseconds=2),
                ZonedDateTime(2000, 1, 2, 3, 4, 5, nanosecond=123456000, tz=UTC.key),
            ),
            param(
                TimeDelta(nanoseconds=2),
                ZonedDateTime(2000, 1, 2, 3, 4, 5, nanosecond=123456788, tz=UTC.key),
            ),
        ],
    )
    def test_date_time_intraday(self, *, delta: Delta, expected: ZonedDateTime) -> None:
        now = ZonedDateTime(2000, 1, 2, 3, 4, 5, nanosecond=123456789, tz=UTC.key)
        result = round_date_or_date_time(now, delta, mode="floor")
        assert result.exact_eq(expected)

    @mark.parametrize(
        ("date_time", "expected"),
        [
            param(
                ZonedDateTime(2000, 1, 1, 2, 3, 4, nanosecond=123456789, tz=UTC.key),
                ZonedDateTime(1999, 12, 31, tz=UTC.key),
            ),
            param(
                ZonedDateTime(2000, 1, 1, tz=UTC.key),
                ZonedDateTime(1999, 12, 31, tz=UTC.key),
            ),
            param(
                ZonedDateTime(2000, 1, 2, tz=UTC.key),
                ZonedDateTime(2000, 1, 2, tz=UTC.key),
            ),
        ],
    )
    def test_date_time_daily(
        self, *, date_time: ZonedDateTime, expected: ZonedDateTime
    ) -> None:
        result = round_date_or_date_time(date_time, DateDelta(days=2), mode="floor")
        assert result.exact_eq(expected)

    @mark.parametrize(
        "delta",
        [
            param(TimeDelta(hours=5)),
            param(TimeDelta(minutes=7)),
            param(TimeDelta(seconds=7)),
            param(TimeDelta(milliseconds=3)),
            param(TimeDelta(microseconds=3)),
            param(TimeDelta(nanoseconds=3)),
        ],
    )
    def test_error_increment(self, *, delta: TimeDelta) -> None:
        with raises(
            _RoundDateOrDateTimeIncrementError,
            match=r"Duration PT.* increment must be a proper divisor of \d+; got \d+",
        ):
            _ = round_date_or_date_time(TODAY_UTC, delta)

    def test_error_invalid(self) -> None:
        with raises(
            _RoundDateOrDateTimeInvalidDurationError,
            match="Duration must be valid; got P1M",
        ):
            _ = round_date_or_date_time(TODAY_UTC, MONTH)

    def test_error_date_with_weekday(self) -> None:
        with raises(
            _RoundDateOrDateTimeDateWithWeekdayError,
            match=r"Daily rounding must not be given a weekday; got Weekday\.MONDAY",
        ):
            _ = round_date_or_date_time(TODAY_UTC, DAY, weekday=Weekday.MONDAY)

    def test_error_date_with_intraday_delta(self) -> None:
        with raises(
            _RoundDateOrDateTimeDateWithIntradayDeltaError,
            match="Dates must not be given intraday durations; got .* and PT1S",
        ):
            _ = round_date_or_date_time(TODAY_UTC, SECOND)

    def test_error_date_time_intra_day_with_weekday(self) -> None:
        with raises(
            _RoundDateOrDateTimeDateTimeIntraDayWithWeekdayError,
            match=r"Date-times and intraday rounding must not be given a weekday; got .*, PT1S and Weekday\.MONDAY",
        ):
            _ = round_date_or_date_time(NOW_UTC, SECOND, weekday=Weekday.MONDAY)


class TestToDate:
    @given(date=dates())
    def test_date(self, *, date: Date) -> None:
        assert to_date(date=date) == date

    @given(date=none() | sentinels())
    def test_none_or_sentinel(self, *, date: None | Sentinel) -> None:
        assert to_date(date=date) is date

    @given(date1=dates(), date2=dates())
    def test_replace_non_sentinel(self, *, date1: Date, date2: Date) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            date: Date = field(default_factory=get_today)

            def replace(self, *, date: MaybeCallableDate | Sentinel = sentinel) -> Self:
                return replace_non_sentinel(self, date=to_date(date=date))

        obj = Example(date=date1)
        assert obj.date == date1
        assert obj.replace().date == date1
        assert obj.replace(date=date2).date == date2
        assert obj.replace(date=get_today).date == get_today()

    @given(date=dates())
    def test_callable(self, *, date: Date) -> None:
        assert to_date(date=lambda: date) == date


class TestToDays:
    @given(cls=sampled_from([DateDelta, DateTimeDelta]), days=integers())
    def test_date_or_date_time_delta(
        self, *, cls: type[DateOrDateTimeDelta], days: int
    ) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="days out of range"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = cls(days=days)
        assert to_days(delta) == days

    @given(days=integers())
    def test_time_delta(self, *, days: int) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="hours out of range"),
            assume_does_not_raise(OverflowError, match="int too big to convert"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = TimeDelta(hours=24 * days)
        assert to_days(delta) == days

    @mark.parametrize(
        "delta", [param(DateDelta(months=1)), param(DateTimeDelta(months=1))]
    )
    def test_error_months(self, *, delta: DateOrDateTimeDelta) -> None:
        with raises(_ToDaysMonthsError, match="Delta must not contain months; got 1"):
            _ = to_days(delta)

    @mark.parametrize(
        "delta", [param(TimeDelta(nanoseconds=1)), param(DateTimeDelta(nanoseconds=1))]
    )
    def test_error_nanoseconds(self, *, delta: TimeOrDateTimeDelta) -> None:
        with raises(
            _ToDaysNanosecondsError,
            match="Delta must not contain extra nanoseconds; got .*",
        ):
            _ = to_days(delta)


class TestToHours:
    @given(days=integers())
    def test_date_delta(self, *, days: int) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="days out of range"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = DateDelta(days=days)
        assert to_hours(delta) == (24 * days)

    @given(cls=sampled_from([TimeDelta, DateTimeDelta]), hours=integers())
    def test_time_or_date_time_delta(
        self, *, cls: type[TimeOrDateTimeDelta], hours: int
    ) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="hours out of range"),
            assume_does_not_raise(OverflowError, match="int too big to convert"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = cls(hours=hours)
        assert to_hours(delta) == hours

    @mark.parametrize(
        "delta", [param(DateDelta(months=1)), param(DateTimeDelta(months=1))]
    )
    def test_error_months(self, *, delta: DateOrDateTimeDelta) -> None:
        with raises(_ToHoursMonthsError, match="Delta must not contain months; got 1"):
            _ = to_hours(delta)

    @mark.parametrize(
        "delta", [param(TimeDelta(nanoseconds=1)), param(DateTimeDelta(nanoseconds=1))]
    )
    def test_error_nanoseconds(self, *, delta: TimeOrDateTimeDelta) -> None:
        with raises(
            _ToHoursNanosecondsError,
            match="Delta must not contain extra nanoseconds; got .*",
        ):
            _ = to_hours(delta)


class TestToLocalPlain:
    @given(date_time=zoned_datetimes())
    def test_main(self, *, date_time: ZonedDateTime) -> None:
        result = to_local_plain(date_time)
        assert isinstance(result, PlainDateTime)


class TestToMicroseconds:
    @given(days=integers())
    def test_date_delta(self, *, days: int) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="days out of range"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = DateDelta(days=days)
        assert to_microseconds(delta) == (24 * 60 * 60 * int(1e6) * days)

    @given(cls=sampled_from([TimeDelta, DateTimeDelta]), microseconds=integers())
    def test_time_or_date_time_delta(
        self, *, cls: type[TimeOrDateTimeDelta], microseconds: int
    ) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="microseconds out of range"),
            assume_does_not_raise(OverflowError, match="int too big to convert"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = cls(microseconds=microseconds)
        assert to_microseconds(delta) == microseconds

    @mark.parametrize(
        "delta", [param(DateDelta(months=1)), param(DateTimeDelta(months=1))]
    )
    def test_error_months(self, *, delta: DateOrDateTimeDelta) -> None:
        with raises(
            _ToMicrosecondsMonthsError, match="Delta must not contain months; got 1"
        ):
            _ = to_microseconds(delta)

    @mark.parametrize(
        "delta", [param(TimeDelta(nanoseconds=1)), param(DateTimeDelta(nanoseconds=1))]
    )
    def test_error_nanoseconds(self, *, delta: TimeOrDateTimeDelta) -> None:
        with raises(
            _ToMicrosecondsNanosecondsError,
            match="Delta must not contain extra nanoseconds; got .*",
        ):
            _ = to_microseconds(delta)


class TestToMilliseconds:
    @given(days=integers())
    def test_date_delta(self, *, days: int) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="days out of range"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = DateDelta(days=days)
        assert to_milliseconds(delta) == (24 * 60 * 60 * int(1e3) * days)

    @given(cls=sampled_from([TimeDelta, DateTimeDelta]), milliseconds=integers())
    def test_time_or_date_time_delta(
        self, *, cls: type[TimeOrDateTimeDelta], milliseconds: int
    ) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="milliseconds out of range"),
            assume_does_not_raise(OverflowError, match="int too big to convert"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = cls(milliseconds=milliseconds)
        assert to_milliseconds(delta) == milliseconds

    @mark.parametrize(
        "delta", [param(DateDelta(months=1)), param(DateTimeDelta(months=1))]
    )
    def test_error_months(self, *, delta: DateOrDateTimeDelta) -> None:
        with raises(
            _ToMillisecondsMonthsError, match="Delta must not contain months; got 1"
        ):
            _ = to_milliseconds(delta)

    @mark.parametrize(
        "delta", [param(TimeDelta(nanoseconds=1)), param(DateTimeDelta(nanoseconds=1))]
    )
    def test_error_nanoseconds(self, *, delta: TimeOrDateTimeDelta) -> None:
        with raises(
            _ToMillisecondsNanosecondsError,
            match="Delta must not contain extra nanoseconds; got .*",
        ):
            _ = to_milliseconds(delta)


class TestToMinutes:
    @given(days=integers())
    def test_date_delta(self, *, days: int) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="days out of range"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = DateDelta(days=days)
        assert to_minutes(delta) == (24 * 60 * days)

    @given(cls=sampled_from([TimeDelta, DateTimeDelta]), minutes=integers())
    def test_time_or_date_time_delta(
        self, *, cls: type[TimeOrDateTimeDelta], minutes: int
    ) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="minutes out of range"),
            assume_does_not_raise(OverflowError, match="int too big to convert"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = cls(minutes=minutes)
        assert to_minutes(delta) == minutes

    @mark.parametrize(
        "delta", [param(DateDelta(months=1)), param(DateTimeDelta(months=1))]
    )
    def test_error_months(self, *, delta: DateOrDateTimeDelta) -> None:
        with raises(
            _ToMinutesMonthsError, match="Delta must not contain months; got 1"
        ):
            _ = to_minutes(delta)

    @mark.parametrize(
        "delta", [param(TimeDelta(nanoseconds=1)), param(DateTimeDelta(nanoseconds=1))]
    )
    def test_error_nanoseconds(self, *, delta: TimeOrDateTimeDelta) -> None:
        with raises(
            _ToMinutesNanosecondsError,
            match="Delta must not contain extra nanoseconds; got .*",
        ):
            _ = to_minutes(delta)


class TestToMonths:
    @given(cls=sampled_from([DateDelta, DateTimeDelta]), months=integers())
    def test_main(self, *, cls: type[DateOrDateTimeDelta], months: int) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="months out of range"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = cls(months=months)
        assert to_months(delta) == months

    @mark.parametrize("delta", [param(DateDelta(days=1)), param(DateTimeDelta(days=1))])
    def test_error_days(self, *, delta: DateOrDateTimeDelta) -> None:
        with raises(_ToMonthsDaysError, match="Delta must not contain days; got 1"):
            _ = to_months(delta)

    def test_error_date_time_delta_time(self) -> None:
        delta = DateTimeDelta(nanoseconds=1)
        with raises(
            _ToMonthsTimeError, match="Delta must not contain a time part; got .*"
        ):
            _ = to_months(delta)


class TestToMonthsAndDays:
    @given(
        cls=sampled_from([DateDelta, DateTimeDelta]), months=integers(), days=integers()
    )
    def test_main(
        self, *, cls: type[DateOrDateTimeDelta], months: int, days: int
    ) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="Mixed sign in Date(Time)?Delta"),
            assume_does_not_raise(ValueError, match="months out of range"),
            assume_does_not_raise(ValueError, match="days out of range"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = cls(months=months, days=days)
        assert to_months_and_days(delta) == (months, days)

    def test_error_date_time_delta_time(self) -> None:
        delta = DateTimeDelta(nanoseconds=1)
        with raises(
            ToMonthsAndDaysError, match="Delta must not contain a time part; got .*"
        ):
            _ = to_months_and_days(delta)


class TestToNanoseconds:
    @given(func=sampled_from([to_time_delta, to_date_time_delta]), nanos=integers())
    def test_main(
        self, *, func: Callable[[int], TimeOrDateTimeDelta], nanos: int
    ) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="TimeDelta out of range"),
            assume_does_not_raise(ValueError, match="total days out of range"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
            assume_does_not_raise(OverflowError, match="int too big to convert"),
        ):
            delta = func(nanos)
        assert to_nanoseconds(delta) == nanos

    @mark.parametrize(
        "delta", [param(DateDelta(months=1)), param(DateTimeDelta(months=1))]
    )
    def test_error(self, *, delta: DateOrDateTimeDelta) -> None:
        with raises(ToNanosecondsError, match="Delta must not contain months; got 1"):
            _ = to_nanoseconds(delta)


class TestToPyDateOrDateTime:
    @mark.parametrize(
        ("date_or_date_time", "expected"),
        [
            param(Date(2000, 1, 1), dt.date(2000, 1, 1)),
            param(
                ZonedDateTime(2000, 1, 1, tz=UTC.key),
                dt.datetime(2000, 1, 1, tzinfo=UTC),
            ),
            param(None, None),
        ],
    )
    def test_main(
        self,
        *,
        date_or_date_time: Date | ZonedDateTime | None,
        expected: dt.date | None,
    ) -> None:
        result = to_py_date_or_date_time(date_or_date_time)
        assert result == expected


class TestToPyTimeDelta:
    @mark.parametrize(
        ("delta", "expected"),
        [
            param(DateDelta(days=1), dt.timedelta(days=1)),
            param(TimeDelta(microseconds=1), dt.timedelta(microseconds=1)),
            param(
                DateTimeDelta(days=1, microseconds=1),
                dt.timedelta(days=1, microseconds=1),
            ),
            param(None, None),
        ],
    )
    def test_main(self, *, delta: Delta | None, expected: dt.timedelta | None) -> None:
        result = to_py_time_delta(delta)
        assert result == expected

    def test_error(self) -> None:
        delta = TimeDelta(nanoseconds=1)
        with raises(
            ToPyTimeDeltaError, match="Time delta must not contain nanoseconds; got 1"
        ):
            _ = to_py_time_delta(delta)


class TestToSeconds:
    @given(days=integers())
    def test_date_delta(self, *, days: int) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="days out of range"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = DateDelta(days=days)
        assert to_seconds(delta) == (24 * 60 * 60 * days)

    @given(cls=sampled_from([TimeDelta, DateTimeDelta]), seconds=integers())
    def test_time_or_date_time_delta(
        self, *, cls: type[TimeOrDateTimeDelta], seconds: int
    ) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="seconds out of range"),
            assume_does_not_raise(OverflowError, match="int too big to convert"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = cls(seconds=seconds)
        assert to_seconds(delta) == seconds

    @mark.parametrize(
        "delta", [param(DateDelta(months=1)), param(DateTimeDelta(months=1))]
    )
    def test_error_months(self, *, delta: DateOrDateTimeDelta) -> None:
        with raises(
            _ToSecondsMonthsError, match="Delta must not contain months; got 1"
        ):
            _ = to_seconds(delta)

    @mark.parametrize(
        "delta", [param(TimeDelta(nanoseconds=1)), param(DateTimeDelta(nanoseconds=1))]
    )
    def test_error_nanoseconds(self, *, delta: TimeOrDateTimeDelta) -> None:
        with raises(
            _ToSecondsNanosecondsError,
            match="Delta must not contain extra nanoseconds; got .*",
        ):
            _ = to_seconds(delta)


class TestToWeeks:
    @given(cls=sampled_from([DateDelta, DateTimeDelta]), weeks=integers())
    def test_date_or_date_time_delta(
        self, *, cls: type[DateOrDateTimeDelta], weeks: int
    ) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="days out of range"),
            assume_does_not_raise(ValueError, match="weeks out of range"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = cls(weeks=weeks)
        assert to_weeks(delta) == weeks

    @given(weeks=integers())
    def test_time_delta(self, *, weeks: int) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="hours out of range"),
            assume_does_not_raise(OverflowError, match="int too big to convert"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = TimeDelta(hours=7 * 24 * weeks)
        assert to_weeks(delta) == weeks

    @mark.parametrize(
        "delta", [param(DateDelta(months=1)), param(DateTimeDelta(months=1))]
    )
    def test_error_months(self, *, delta: DateOrDateTimeDelta) -> None:
        with raises(_ToWeeksMonthsError, match="Delta must not contain months; got 1"):
            _ = to_weeks(delta)

    @mark.parametrize("delta", [param(DateDelta(days=8)), param(DateTimeDelta(days=8))])
    def test_error_days(self, *, delta: DateOrDateTimeDelta) -> None:
        with raises(
            _ToWeeksDaysError, match="Delta must not contain extra days; got 1"
        ):
            _ = to_weeks(delta)

    @mark.parametrize(
        "delta", [param(TimeDelta(nanoseconds=1)), param(DateTimeDelta(nanoseconds=1))]
    )
    def test_error_nanoseconds(self, *, delta: TimeOrDateTimeDelta) -> None:
        with raises(
            _ToWeeksNanosecondsError,
            match="Delta must not contain extra nanoseconds; got .*",
        ):
            _ = to_weeks(delta)


class TestToYears:
    @given(cls=sampled_from([DateDelta, DateTimeDelta]), years=integers())
    def test_main(self, *, cls: type[DateOrDateTimeDelta], years: int) -> None:
        with (
            assume_does_not_raise(ValueError, match="Out of range"),
            assume_does_not_raise(ValueError, match="months out of range"),
            assume_does_not_raise(ValueError, match="years out of range"),
            assume_does_not_raise(
                OverflowError, match="Python int too large to convert to C long"
            ),
        ):
            delta = cls(years=years)
        assert to_years(delta) == years

    @mark.parametrize(
        "delta", [param(DateDelta(months=1)), param(DateTimeDelta(months=1))]
    )
    def test_error_date_delta_months(self, *, delta: DateOrDateTimeDelta) -> None:
        with raises(_ToYearsMonthsError, match="Delta must not contain months; got 1"):
            _ = to_years(delta)

    @mark.parametrize("delta", [param(DateDelta(days=1)), param(DateTimeDelta(days=1))])
    def test_error_date_delta_days(self, *, delta: DateOrDateTimeDelta) -> None:
        with raises(_ToYearsDaysError, match="Delta must not contain days; got 1"):
            _ = to_years(delta)

    def test_error_date_time_delta_time(self) -> None:
        delta = DateTimeDelta(nanoseconds=1)
        with raises(
            _ToYearsTimeError, match="Delta must not contain a time part; got .*"
        ):
            _ = to_years(delta)


class TestToZonedDateTime:
    @given(date_time=zoned_datetimes())
    def test_date_time(self, *, date_time: ZonedDateTime) -> None:
        assert to_zoned_date_time(date_time=date_time) == date_time

    @given(date_time=none() | sentinels())
    def test_none_or_sentinel(self, *, date_time: None | Sentinel) -> None:
        assert to_zoned_date_time(date_time=date_time) is date_time

    @given(date_time1=zoned_datetimes(), date_time2=zoned_datetimes())
    def test_replace_non_sentinel(
        self, *, date_time1: ZonedDateTime, date_time2: ZonedDateTime
    ) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            date_time: ZonedDateTime = field(default_factory=get_now)

            def replace(
                self, *, date_time: MaybeCallableZonedDateTime | Sentinel = sentinel
            ) -> Self:
                return replace_non_sentinel(
                    self, date_time=to_zoned_date_time(date_time=date_time)
                )

        obj = Example(date_time=date_time1)
        assert obj.date_time == date_time1
        assert obj.replace().date_time == date_time1
        assert obj.replace(date_time=date_time2).date_time == date_time2
        assert abs(obj.replace(date_time=get_now).date_time - get_now()) <= SECOND

    @given(date_time=zoned_datetimes())
    def test_callable(self, *, date_time: ZonedDateTime) -> None:
        assert to_zoned_date_time(date_time=lambda: date_time) == date_time


class TestTwoDigitYearMonth:
    def test_parse_common_iso(self) -> None:
        result = two_digit_year_month(0, 1)
        expected = YearMonth(2000, 1)
        assert result == expected


class TestWheneverLogRecord:
    def test_init(self) -> None:
        _ = WheneverLogRecord("name", DEBUG, "pathname", 0, None, None, None)

    def test_get_length(self) -> None:
        assert isinstance(WheneverLogRecord._get_length(), int)

    def test_get_time_zone(self) -> None:
        assert isinstance(WheneverLogRecord._get_time_zone(), ZoneInfo)

    def test_get_time_zone_key(self) -> None:
        assert isinstance(WheneverLogRecord._get_time_zone_key(), str)



from __future__ import annotations

import re
from dataclasses import dataclass
from re import Pattern
from typing import TYPE_CHECKING, assert_never, override

if TYPE_CHECKING:
    from utilities.types import PatternLike


def ensure_pattern(pattern: PatternLike, /, *, flags: int = 0) -> Pattern[str]:
    """Ensure a pattern is returned."""
    match pattern:
        case Pattern():
            return pattern
        case str():
            return re.compile(pattern, flags=flags)
        case _ as never:
            assert_never(never)


##


def extract_group(pattern: PatternLike, text: str, /, *, flags: int = 0) -> str:
    """Extract a group.

    The regex must have 1 capture group, and this must match exactly once.
    """
    pattern_use = ensure_pattern(pattern, flags=flags)
    match pattern_use.groups:
        case 0:
            raise _ExtractGroupNoCaptureGroupsError(pattern=pattern_use, text=text)
        case 1:
            matches: list[str] = pattern_use.findall(text)
            match len(matches):
                case 0:
                    raise _ExtractGroupNoMatchesError(
                        pattern=pattern_use, text=text
                    ) from None
                case 1:
                    return matches[0]
                case _:
                    raise _ExtractGroupMultipleMatchesError(
                        pattern=pattern_use, text=text, matches=matches
                    ) from None
        case _:
            raise _ExtractGroupMultipleCaptureGroupsError(
                pattern=pattern_use, text=text
            )


@dataclass(kw_only=True, slots=True)
class ExtractGroupError(Exception):
    pattern: Pattern[str]
    text: str


@dataclass(kw_only=True, slots=True)
class _ExtractGroupMultipleCaptureGroupsError(ExtractGroupError):
    @override
    def __str__(self) -> str:
        return f"Pattern {self.pattern} must contain exactly one capture group; it had multiple"


@dataclass(kw_only=True, slots=True)
class _ExtractGroupMultipleMatchesError(ExtractGroupError):
    matches: list[str]

    @override
    def __str__(self) -> str:
        return f"Pattern {self.pattern} must match against {self.text} exactly once; matches were {self.matches}"


@dataclass(kw_only=True, slots=True)
class _ExtractGroupNoCaptureGroupsError(ExtractGroupError):
    @override
    def __str__(self) -> str:
        return f"Pattern {self.pattern} must contain exactly one capture group; it had none".format(
            self.pattern
        )


@dataclass(kw_only=True, slots=True)
class _ExtractGroupNoMatchesError(ExtractGroupError):
    @override
    def __str__(self) -> str:
        return f"Pattern {self.pattern} must match against {self.text}"


##


def extract_groups(pattern: PatternLike, text: str, /, *, flags: int = 0) -> list[str]:
    """Extract multiple groups.

    The regex may have any number of capture groups, and they must collectively
    match exactly once.
    """
    pattern_use = ensure_pattern(pattern, flags=flags)
    if (n_groups := pattern_use.groups) == 0:
        raise _ExtractGroupsNoCaptureGroupsError(pattern=pattern_use, text=text)
    matches: list[str] = pattern_use.findall(text)
    match len(matches), n_groups:
        case 0, _:
            raise _ExtractGroupsNoMatchesError(pattern=pattern_use, text=text)
        case 1, 1:
            return matches
        case 1, _:
            return list(matches[0])
        case _:
            raise _ExtractGroupsMultipleMatchesError(
                pattern=pattern_use, text=text, matches=matches
            )


@dataclass(kw_only=True, slots=True)
class ExtractGroupsError(Exception):
    pattern: Pattern[str]
    text: str


@dataclass(kw_only=True, slots=True)
class _ExtractGroupsMultipleMatchesError(ExtractGroupsError):
    matches: list[str]

    @override
    def __str__(self) -> str:
        return f"Pattern {self.pattern} must match against {self.text} exactly once; matches were {self.matches}"


@dataclass(kw_only=True, slots=True)
class _ExtractGroupsNoCaptureGroupsError(ExtractGroupsError):
    pattern: Pattern[str]
    text: str

    @override
    def __str__(self) -> str:
        return f"Pattern {self.pattern} must contain at least one capture group"


@dataclass(kw_only=True, slots=True)
class _ExtractGroupsNoMatchesError(ExtractGroupsError):
    @override
    def __str__(self) -> str:
        return f"Pattern {self.pattern} must match against {self.text}"


__all__ = [
    "ExtractGroupError",
    "ExtractGroupsError",
    "ensure_pattern",
    "extract_group",
    "extract_groups",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from pytest import mark, param

from utilities.inflect import counted_noun

if TYPE_CHECKING:
    from collections.abc import Sized


class TestCountedNoun:
    @mark.parametrize(
        ("obj", "expected"),
        [
            param(0, "0 words"),
            param(1, "1 word"),
            param(2, "2 words"),
            param(3, "3 words"),
            param([], "0 words"),
            param(["a"], "1 word"),
            param(["a", "b"], "2 words"),
            param(["a", "b", "c"], "3 words"),
        ],
    )
    def test_main(self, *, obj: int | Sized, expected: str) -> None:
        result = counted_noun(obj, "word")
        assert result == expected



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, cast, override

from hypothesis import given
from hypothesis.strategies import booleans
from luigi import BoolParameter, Task
from pytest import mark, param

from utilities.hypothesis import namespace_mixins, temp_paths, zoned_datetimes
from utilities.luigi import (
    ExternalFile,
    ExternalTask,
    PathTarget,
    ZonedDateTimeParameter,
    _ExternalTaskDummyTarget,
    build,
)

if TYPE_CHECKING:
    from whenever import ZonedDateTime


class TestBuild:
    @given(namespace_mixin=namespace_mixins())
    def test_main(self, *, namespace_mixin: Any) -> None:
        class Example(namespace_mixin, Task): ...

        _ = build([Example()], local_scheduler=True)


class TestDateTimeParameter:
    @given(datetime=zoned_datetimes())
    @mark.parametrize("type_", [param("datetime"), param("str")])
    def test_main(
        self, *, datetime: ZonedDateTime, type_: Literal["datetime", "str"]
    ) -> None:
        param = ZonedDateTimeParameter()
        match type_:
            case "datetime":
                input_ = datetime
            case "str":
                input_ = datetime.format_common_iso()
        norm = param.normalize(input_)
        assert param.parse(param.serialize(norm)) == norm


class TestExternalFile:
    @given(namespace_mixin=namespace_mixins(), root=temp_paths())
    def test_main(self, *, namespace_mixin: Any, root: Path) -> None:
        path = root.joinpath("file")

        class Example(namespace_mixin, ExternalFile): ...

        task = Example(path)
        assert not task.exists()
        path.touch()
        assert task.exists()


class TestExternalTask:
    @given(namespace_mixin=namespace_mixins(), is_complete=booleans())
    def test_main(self, *, namespace_mixin: Any, is_complete: bool) -> None:
        class Example(namespace_mixin, ExternalTask):
            is_complete: bool = cast("bool", BoolParameter())

            @override
            def exists(self) -> bool:
                return self.is_complete

        task = Example(is_complete=is_complete)
        result = task.exists()
        assert result is is_complete
        assert isinstance(task.output(), _ExternalTaskDummyTarget)
        result2 = task.output().exists()
        assert result2 is is_complete


class TestPathTarget:
    def test_main(self, *, tmp_path: Path) -> None:
        target = PathTarget(path := tmp_path.joinpath("file"))
        assert isinstance(target.path, Path)
        assert not target.exists()
        path.touch()
        assert target.exists()



from __future__ import annotations

from asyncio import get_running_loop
from typing import TYPE_CHECKING

from aiolimiter import AsyncLimiter

if TYPE_CHECKING:
    from collections.abc import Hashable

_LIMITERS: dict[tuple[int, Hashable], AsyncLimiter] = {}


def get_async_limiter(key: Hashable, /, *, rate: float = 1.0) -> AsyncLimiter:
    """Get a loop-aware rate limiter."""
    id_ = id(get_running_loop())
    full = (id_, key)
    try:
        return _LIMITERS[full]
    except KeyError:
        limiter = _LIMITERS[full] = AsyncLimiter(1.0, time_period=rate)
        return limiter


__all__ = ["get_async_limiter"]



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from hypothesis import HealthCheck, given, settings
from hypothesis.strategies import sampled_from
from polars import int_range

from tests.test_typing_funcs.with_future import (
    DataClassFutureInt,
    DataClassFutureLiteral,
    DataClassFutureNestedOuterFirstInner,
    DataClassFutureNestedOuterFirstOuter,
)

if TYPE_CHECKING:
    from utilities.pytest_regressions import (
        OrjsonRegressionFixture,
        PolarsRegressionFixture,
    )


class TestMultipleRegressionFixtures:
    def test_main(
        self,
        *,
        orjson_regression: OrjsonRegressionFixture,
        polars_regression: PolarsRegressionFixture,
    ) -> None:
        obj = DataClassFutureInt(int_=0)
        orjson_regression.check(obj, suffix="obj")
        series = int_range(end=10, eager=True).alias("value")
        polars_regression.check(series, suffix="series")


class TestPolarsRegressionFixture:
    def test_dataframe(self, *, polars_regression: PolarsRegressionFixture) -> None:
        df = int_range(end=10, eager=True).alias("value").to_frame()
        polars_regression.check(df)

    def test_series(self, *, polars_regression: PolarsRegressionFixture) -> None:
        series = int_range(end=10, eager=True).alias("value")
        polars_regression.check(series)


class TestOrjsonRegressionFixture:
    def test_dataclass_nested(
        self, *, orjson_regression: OrjsonRegressionFixture
    ) -> None:
        obj = DataClassFutureNestedOuterFirstOuter(
            inner=DataClassFutureNestedOuterFirstInner(int_=0)
        )
        orjson_regression.check(obj)

    def test_dataclass_int(self, *, orjson_regression: OrjsonRegressionFixture) -> None:
        obj = DataClassFutureInt(int_=0)
        orjson_regression.check(obj)

    @given(truth=sampled_from(["true", "false"]))
    @settings(suppress_health_check={HealthCheck.function_scoped_fixture})
    def test_dataclass_literal(
        self,
        *,
        truth: Literal["true", "false"],
        orjson_regression: OrjsonRegressionFixture,
    ) -> None:
        obj = DataClassFutureLiteral(truth=truth)
        orjson_regression.check(obj, suffix=truth)



from __future__ import annotations

from math import isfinite, isnan

from utilities.psutil import MemoryUsage


class TestMemoryUsage:
    def test_main(self) -> None:
        memory = MemoryUsage.new()
        assert memory.virtual_total >= 0
        assert memory.virtual_total_mb >= 0
        assert memory.virtual_total >= 0
        assert memory.virtual_total_mb >= 0
        assert (
            isfinite(memory.virtual_pct) and (0.0 <= memory.virtual_pct <= 1.0)
        ) or isnan(memory.virtual_pct)
        assert memory.swap_total >= 0
        assert memory.swap_total_mb >= 0
        assert memory.swap_total >= 0
        assert memory.swap_total_mb >= 0
        assert (isfinite(memory.swap_pct) and (0.0 <= memory.swap_pct <= 1.0)) or isnan(
            memory.swap_pct
        )
